/*
   netrik -- The ANTRIK Internet Viewer
   Copyright (C) Olaf D. Buddenhagen AKA antrik, et al (see AUTHORS)
   Published under the GNU GPL; see LICENSE for details.
*/
/*
 * load.c -- load file from ressource
 *
 * (C) 2001, 2002 antrik
 *     2001, 2002 Patrice Neff
 *
 * Open a file, HTTP connection etc, and then read data blockwise.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "config.h"

#include "cfg.h"
#include "debug.h"
#include "load.h"
#include "http.h"

#ifdef DEBUG
   #define BUF_SIZE 16
#else
   #define BUF_SIZE 4096
#endif

static void init_wget(struct Resource *res);    /* prepare reading HTTP resource via wget */

/* prepare reading HTTP resource via wget;
 * calls wget in background, using its standard output as our input stream */
static void init_wget(res)
struct Resource	*res;
{
   char	cmd[strlen(res->url->full_url)+sizeof(WGET_CMD)];

   res->type=RES_PIPE;

   /* security workaround: prevent URL from escaping the shell quote (this is a hack...) */
   {
      char	*chr;
      for(chr=res->url->full_url; *chr; ++chr)
	 if(*chr=='\'')
	    *chr=' ';
   }

   sprintf(cmd, WGET_CMD, res->url->full_url);    /* create command line for wget */
   res->handle.stream=popen(cmd, "r");    /* call in background, get output as pipe */
   if(res->handle.stream==NULL) {
      fprintf(stderr, "error executing wget\n");
      if(res->url->proto.type==PT_HTTP)
	 fprintf(stderr, "(try using --builtin-http)\n");
      exit(1);
   }
}

/*
 * build new URL, and prepare addressed resource for reading data
 * (create descriptor, open file/HTTP connection/wget pipe, alloc input buffer)
 */
struct Resource *init_load(base, url, form)
const struct Url	*base;    /* main URL is relative to this one */
const char		*url;    /* main URL, to be merged with "base" */
const struct Item	*form;    /* form item of a form to submit */
{
   struct Resource	*res;    /* opened ressource */
   struct Url		*base_url=(struct Url *)base;    /* avoid warnings about modifying "base" */

   /* alloc resource descriptor */
   res=malloc(sizeof(struct Resource));
   if(res==NULL) {
      fprintf(stderr, "memory allocation failure while opening resource\n");
      exit(1);
   }

   /* alloc input buffer */
   res->buf=malloc(sizeof(char[BUF_SIZE]));
   if(res->buf==NULL) {
      fprintf(stderr, "memory allocation failure while opening resource\n");
      exit(1);
   }
   res->buf_ptr=res->buf_end=res->buf;    /* empty */

   /* open file/connection/pipe */

   if(url!=NULL) {    /* not already a split/merged URL (from history) */
      res->url=merge_urls(base_url, url);
      if(res->url->proto.type==PT_INTERNAL) {    /* couldn't merge URLs */
	 res->type=RES_FAIL;
	 return res;
      }
      if(strcmp(url, "-")==0)    /* load from stdin */
	 res->url->proto.type=PT_INTERNAL;
   } else    /* already split URL */
      res->url=base_url;

   switch(res->url->proto.type) {
      case PT_INTERNAL:    /* use stdin */
	 if(!res->url->absolute) {
	    fprintf(stderr, "error: can't follow relative link from stdin\n");
	    exit(2);
	 }

	 /* reopen stdin; use for data input */
	 res->handle.stream=fdopen(dup(0), "r");
	 if(res->handle.stream==NULL) {
	    fprintf(stderr, "can't read stdin (reopen failed)\n");
	    exit(1);
	 }

	 /* use stderr as stdin */
	 if(dup2(2, 0)<0) {
	    fprintf(stderr, "can't reopen stderr for reading; don't know how to get interactive user input\n");
	    exit(1);
	 }

	 res->type=RES_STDIN;
	 break;

      case PT_FILE:
	 DMSG(("loading file: %s\n", res->url->path));
	 res->handle.stream=fopen(res->url->path, "r");
	 if(res->handle.stream==NULL) {    /* open failed */
	    fprintf(stderr, "can't open file %s\n", url);
	    res->type=RES_FAIL;
	    res->url->proto.type=PT_INTERNAL;    /* don't keep in history */
	    return res;
	 }
	 res->type=RES_FILE;
	 break;

      case PT_UNKNOWN:    /* assume local file */
	 DMSG(("loading file: %s\n", res->url->path));
	 res->handle.stream=fopen(res->url->path, "r");
	 if(res->handle.stream!=NULL) {    /* could open local file */
	    res->type=RES_FILE;
	    res->url->proto.type=PT_FILE;
	    break;
	 } else {    /* failed to open local file */
	    if(res->url->path[0]!='/') {    /* may also be incomplete HTTP URL */
	       char	x_url[sizeof("http://")+strlen(url)];    /* URL string with prepended protocol specification */

	       DMSG(("can't open local file\ntrying HTTP...\n"));
	       sprintf(x_url, "http://%s", url);

	       free_url(res->url);
	       res->url=merge_urls(base_url, x_url);    /* need to parse again */
	       if(res->url->proto.type==PT_INTERNAL) {    /* couldn't merge URLs */
		  res->type=RES_FAIL;
		  return res;
	       }
	       /* fallthrough */
	    } else {    /* can't be HTTP */
	       fprintf(stderr, "can't open file %s\n", url);
	       res->type=RES_FAIL;
	       res->url->proto.type=PT_INTERNAL;    /* don't keep in history */
	       return res;
	    }
	 }

      case PT_HTTP:
	 if(cfg.wget)
	    init_wget(res);
	 else
	    http_init_load(res, form);
	 break;

      case PT_FTP:
	 init_wget(res);
   }    /* switch proto.type */

   return res;
}

/* read data block from ressource into buffer */
void load(res)
struct Resource	*res;
{
   int chars_read;

   if(res->type==RES_FAIL)
      chars_read=0;
/* Patrice, antrik --> */
   else if(res->type == RES_HTTP) {
      chars_read=read(res->handle.http->socket, res->buf, BUF_SIZE);
      if(chars_read<0) {
         fprintf(stderr, "error while reading data from HTTP connection\n");
	 res->type=RES_FAIL;
	 chars_read=0;    /* return EOF */
      }
   } else {
/* <-- Patrice, antrik */
      chars_read=fread(res->buf, sizeof(char), BUF_SIZE, res->handle.stream);
      if(ferror(res->handle.stream)) {
         fprintf(stderr, "error while reading from file\n");
	 res->type=RES_FAIL;
	 chars_read=0;    /* return EOF */
      }
   }
   
   res->buf_end=res->buf+chars_read;
}

/* tidy up after loading a file;
 * closes stream and then frees memory allocated for the buffer and the resource
 * handle */
void uninit_load(res)
struct Resource *res;
{
   if(res->type==RES_PIPE) {
      if(pclose(res->handle.stream)!=0) {
	 fprintf(stderr, "error loading HTTP page\n");
	 exit(1);
      }
   } else if(res->type==RES_FILE || res->type==RES_STDIN)
      fclose(res->handle.stream);

   free(res->buf);
   free(res);
}
